##############################################################
## gff3Plotter - GFF3 file plotter                          ##
## Modified: 28/07/2005                                     ##
## Requires: GDD                                            ##
## Parts of code based on 'tilingArray', 'geneplotter'      ##
## (see notes)                                              ##
## Copyright: Oleg Sklyar  (*)  osklyar@ebi.ac.uk - 2005    ##
##            (structure and core code)                     ##
##            Sophie Grosz (**) grosz@embl.de     - 2005    ##
##            (image maps, coco profile, bug fixes)         ##
## *)  European Bioinformatics Institute                    ##
##     Hinxton, Cambs, CB10 1SD, UK                         ##
## **) European Molecular Biology Laboratory (EMBL)         ##
##     Heidelberg, Germany                                  ##
##                                                          ##
##     LGPL                                                 ##
##############################################################

def.h <- function() {
   h = c(0.15, 0.15, 0.1, 0.02, 0.6, 0.4, 0.15, 0.2, 0.2, 0.05)
   names(h) <- c("gap", "text", "coco", "cocoSep", "curve", "gene", "expro", "exproSep", "axis", "ticks")
   return(h)
}

gffplot.getColours <- function(scores, treshold = 1, baseCols) {
   ## >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
   ## sub-Function to create color ramps .............................
   ## ADAPTED: from 'geneplotter' colorRamp function. In the new release of BioC
   ## colorRamp was moved out of 'geneplotter' into the base also changing the
   ## interface and behaviour!
   localColorRamp <- function(palette, bias = 1) {
      coord <- as.data.frame(t(col2rgb(palette)) / 255)
      x <- seq(0, 1, length = length(palette))
      r <- approxfun(x, coord$red)
      g <- approxfun(x, coord$green)
      b <- approxfun(x, coord$blue)
      function(x, min = 0, max = 1) {
         if (any(x < min | x > max))
            stop("out of range")
         z <- ((x - min) / (max - min)) ^ bias
         rgb(r(z), g(z), b(z))
      }
   }
   ## <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
   if(length(baseCols) < 3)
      baseCols = c(baseCols, "blue", "grey", "red")
   getColorFUN = localColorRamp(baseCols[1:3])
   maxDeviation = max(c(abs(max(scores) - treshold), abs(min(scores) - treshold)))
   if(maxDeviation == 0)
      maxDeviation = 1
   scoreRange = c(treshold - maxDeviation, treshold + maxDeviation)
   cols = (0.5 * (scores - scoreRange[[1]]) / maxDeviation)
   cols[cols < 0] = 0
   cols[cols > 1] = 1
   cols = getColorFUN(cols)
   invisible(cols)
}

## calculates axis label tick positions .................
## ADAPTED: from 'tilingArray' alongChromTicks function
getTickCoords <- function(range) {
   logDelta = log(0.2 * (range[[2]] - range[[1]]), 10)
   logDeltaFloor = floor(logDelta)
   if(logDelta - logDeltaFloor > log(5, 10))
      logDeltaFloor = logDeltaFloor +  log(5, 10)
   delta = round(10 ^ logDeltaFloor)
   startInd = floor(range[[1]] / delta)
   endInd = ceiling(range[[2]] / delta)
   ticks = seq(startInd, endInd) * delta
   ticks = ticks[ticks >= range[[1]] & ticks <= range[[2]]]
   invisible(ticks)
}

gffplot.axis <- function(top, xrange, yrange, h) {
   xdelta = xrange$max - xrange$min
   axispos = top - 0.7 * h$axis
   ticks = getTickCoords(as.numeric(xrange))
   segments(ticks, yrange$max, ticks, yrange$min, col = "lightgray", lwd = 1)
   arrows(xrange$min - xdelta * 0.05, axispos, xrange$max + xdelta * 0.05, axispos, length = 0.15, angle = 12, code = 2, col = "#132E73", xpd = NULL)
   segments(ticks, axispos + 0.5 * h$ticks, ticks, axispos - 0.5 * h$ticks, col = "#132E73")
   tickLabels = as.character(ticks)
   index = ticks >= 1000 & ticks < 1000000
   tickLabels[index] = paste(ticks[index] * 0.001, "kb")
   index = ticks >= 1000000
   tickLabels[index] = paste(ticks[index] * 0.000001, "Mb")
   text(x = ticks, y = axispos + 1.5 * h$ticks, labels = tickLabels, col = "#132E73", adj = c(0.5, 0), font = 2)
   names(ticks) = tickLabels
   invisible(ticks)
}

gffplot.coco <- function(coco, xrange, coco.tresholds, top, h, nlevels, def.cols = c("#51821B", "#B12525", "darkblue"), do.ramp = TRUE) { #c("#FF6347", "#CCCCCC", "darkblue")) {
   ncoco = length(coco)
   for (i in 1:ncoco) {
      cocoi = coco[[i]]
      ncocoi = length(cocoi$START)
      if (ncocoi > 0) {
         # setting vertical positions
         ## will be repeated as many times as needed
         cocoi$START[cocoi$START < xrange$min] = xrange$min
         cocoi$END[cocoi$END < xrange$min] = xrange$min
         cocoi$START[cocoi$START > xrange$max] = xrange$max
         cocoi$END[cocoi$END > xrange$max] = xrange$max
         topsi = numeric(nlevels)
         topsi[] = top - (i - 1) * ((h$coco + h$cocoSep) * nlevels + h$gap)
         topsi[] = topsi[] - (h$coco + h$cocoSep) * (1:nlevels - 1)
         botsi = topsi - h$coco
         # setting colors
         if (do.ramp) {
            cols = gffplot.getColours(cocoi$SCORE, coco.tresholds[[i]], c(def.cols[[2]], "grey", def.cols[[1]]))
         } else {
            cols = character(ncocoi)
            cols[cocoi$SCORE >  coco.tresholds[[i]]] = def.cols[[1]]
            cols[cocoi$SCORE <= coco.tresholds[[i]]] = def.cols[[2]]
         }
         cols[cocoi$sticky]                       = def.cols[[3]]
         # plot coco rectangles
         rect(cocoi$START, topsi, cocoi$END, botsi, col = cols, border = cols) #FALSE, lwd = 0)
         # output labels
         widths = (cocoi$END - cocoi$START) / (xrange$max - xrange$min)
         labels = cocoi$SEQ_ID
         index = 0.8 * strwidth(labels, units = "figure") > widths
         labels[index] = ""
         text(x = 0.5 * (cocoi$START + cocoi$END), y = 0.5 * (topsi + botsi), labels = labels, col = "black", adj = c(0.5, 0.5), cex = 0.8 * par()$cex)
      }
   }
   invisible(NULL)
}

gffplot.cocoCurve <- function(coco1, xrange, coco.tresholds, top, h, def.cols = c("#51821B", "#B12525", "darkblue")) {
   ncoco1 = length(coco1$START)
   # bar plot
   if (ncoco1 > 1) {
      coco1$START[coco1$START < xrange$min] = xrange$min
      coco1$END[coco1$END < xrange$min] = xrange$min
      coco1$START[coco1$START > xrange$max] = xrange$max
      coco1$END[coco1$END > xrange$max] = xrange$max
      # setting colors
      rampCols = c(def.cols[[2]], "grey", def.cols[[1]])
      cols = gffplot.getColours(coco1$SCORE, coco.tresholds[[1]], rampCols)
      cols[coco1$sticky]                       = def.cols[[3]]
      # calculating y coordinates
      miny = min(coco1$SCORE)
      maxy = max(coco1$SCORE)
      if (miny > coco.tresholds[[1]])
         miny = coco.tresholds[[1]]
      if (maxy < coco.tresholds[[1]])
         maxy = coco.tresholds[[1]]
      if (maxy - miny < 0.000000001)
         invisible(NULL)
      xdelta = xrange$max - xrange$min
      coord = (coco1$SCORE - miny) / (maxy - miny) * h$curve + top - h$curve
      zero = (coco.tresholds[[1]] - miny) / (maxy - miny) * h$curve + top - h$curve
      rect(coco1$START, coord, coco1$END, zero, col = cols, border = cols) #FALSE, lwd = 0)
      text(xrange$min - 0.01 * xdelta, top, labels = maxy, adj = c(1, 0.5), font = 3)
      text(xrange$min - 0.01 * xdelta, top - h$curve, labels = miny, adj = c(1, 0.5), font = 3)
      text(xrange$min + 0.01 * xdelta, zero - h$ticks, labels = names(coco.tresholds)[[1]], adj = c(0, 1), font = 2)
      arrows(xrange$min - 0.02 * xdelta, zero, xrange$max + 0.03 * xdelta, zero, length = 0.1, angle = 12, code = 2, col = "darkblue", xpd = NULL)
      arrows(xrange$min, top - h$curve - 2 * h$ticks, xrange$min, top + h$gap, length = 0.1, angle = 12, code = 2, col = "darkblue", xpd = NULL)
   }
   invisible(NULL)
}

gffplot.genesAndExons <- function(gene, exon, insitu, strand, xrange, par.devstage, par.anatomy, top, h, def.cols = c("orange", "tomato", "blue", "lightgrey")) {
   ngene = length(gene$START)
   if (ngene <= 0)
      invisible(NULL)
   # abline(h = top - 0.5 * h$gene, lwd = 2, col = "darkgrey")
   segments(xrange$min, top - 0.5 * h$gene, xrange$max, top - 0.5 * h$gene, lwd = 2, col = "darkgrey")
   cols = character(ngene)
   cols[] = def.cols[[4]]
   # setting gene colors depending on insitu data
   # it is assumed that several lines can contain complementary information
   if (!is.null(insitu)) {
      ninsitu = length(insitu$START)
      if (ninsitu > 0) {
         # only those of insitu that have muscle in antomy
         anatomyIDs = insitu$SEQ_ID[grep("muscle", insitu$anatomy)]
         # only those that have matching dev stages
         expr = paste(par.devstage, collapse = "|")
         devStageIDs = insitu$SEQ_ID[grep(expr, insitu$devStage)]
         # match anatomy SEQ_ID to devStage SEQ_ID and separate those that match
         index = match(anatomyIDs, devStageIDs)
         bothIDs = anatomyIDs[!is.na(index)]
         anatomy = anatomyIDs[is.na(index)]
         index = match(devStageIDs, bothIDs)
         devStageIDs = devStageIDs[is.na(index)]
         # set color of genes, for which voth are present
         index = match(gene$SEQ_ID, bothIDs)
         cols[!is.na(index)] = def.cols[[1]]
         # set color of genes, for which only 'muscle' is present
         index = match(gene$SEQ_ID, anatomyIDs)
         cols[!is.na(index)] = def.cols[[2]]
         # set color of genes, for which only 'devStage' is present
         index = match(gene$SEQ_ID, devStageIDs)
         cols[!is.na(index)] = def.cols[[3]]
      }
   }
   # keep data that can be changed later but need to be used original
   xdelta = xrange$max - xrange$min
   labelStarts = gene$START
   labelEnds = gene$END
   # change starts < xrange$min to xrange$min
   indexStarts = gene$START < xrange$min
   gene$START[indexStarts] = xrange$min
   gene$END[gene$END < xrange$min] = xrange$min
   # change ends > xrange$max to xrange$max
   indexEnds = gene$END > xrange$max
   gene$END[indexEnds] = xrange$max
   gene$START[gene$START > xrange$max] = xrange$max
   # plot gene rectangles
   rect(gene$START, top, gene$END, top - h$gene, col = cols, border = cols) #FALSE) #"black", lwd = 1)
   # plot open side gene start if there is a gene with start < xrange$min
   if (strand == "+")
      x = c(xrange$max, xrange$max, xrange$max + 0.02 * xdelta)
   else
      x = c(xrange$min, xrange$min, xrange$min - 0.02 * xdelta)
   y = c(top, top - h$gene, top - 0.5 * h$gene)
   polygon(x = x, y = y, col = "#A2ACC9", border = "#A2ACC9")
   # plot exons if present
   if (!is.null(exon)) {
      exon$START[exon$START < xrange$min] = xrange$min
      exon$END[exon$END < xrange$min] = xrange$min
      exon$START[exon$START > xrange$max] = xrange$max
      exon$END[exon$END > xrange$max] = xrange$max
      rect(exon$START, top, exon$END, top - h$gene, col = "#2F2F2F", border = "#2F2F2F") #FALSE, lwd = 0)
   }
   # output gene names
   widths = (gene$END - gene$START) / xdelta
   # index of genes that have too long names
   labels = gene$ID
   index = strwidth(labels, units = "figure") > widths
   labels[index] = ""
   text(x = 0.5 * (gene$START + gene$END), y = top + h$ticks, labels = labels, col = "black", adj = c(0.5, 0), font = 2)
   # output gene coordinates
   # index of genes that have too long coordinate strings
   labels = paste(labelStarts, labelEnds, sep = "   ")
   index = strwidth(labels, units = "figure") > widths
   labelStarts[index] = ""
   text(x = gene$START, y = top - h$gene - h$ticks, labels = labelStarts, col = "black", adj = c(0, 1), font = 3)
   labelEnds[index] = ""
   text(x = gene$END, y = top - h$gene - h$ticks, labels = labelEnds, col = "black", adj = c(1, 1), font = 3)
}

gffplot.expro <- function(expro, strand, xrange, expro.tresholds, top, h, def.cols = c("#2C7BB6", "grey", "#FECC5C")) {
   nexpro = length(expro)
   for (i in 1:nexpro) {
      exproi = expro[[i]]
      exproi = exproi[exproi$STRAND == strand, ]
      nexproi = length(exproi$START)
      if (nexproi > 0) {
         topsi = top - h$text - (i - 1) * (h$expro + h$text + h$exproSep)
         botsi = topsi - h$expro
         xdelta = xrange$max - xrange$min
         segments(xrange$min - 0.1 * xdelta, 0.5 * (topsi + botsi), xrange$max, 0.5 * (topsi + botsi))
         exproi$SCORE[is.na(exproi$SCORE)] = expro.tresholds[[i]]
         cols = gffplot.getColours(exproi$SCORE, expro.tresholds[[i]], def.cols)
         indexStarts = exproi$START < xrange$min
         exproi$START[indexStarts] = xrange$min
         exproi$END[exproi$END < xrange$min] = xrange$min
         indexEnds = exproi$END > xrange$max
         exproi$END[indexEnds] = xrange$max
         exproi$START[exproi$START > xrange$max] = xrange$max
         rect(exproi$START, topsi, exproi$END, botsi, col = cols, border = cols) #FALSE) #"black", lwd = 1)
         #if (strand == "+")
         #   x = c(xrange$max, xrange$max, xrange$max + 0.01 * xdelta)
         #else
         #   x = c(xrange$min, xrange$min, xrange$min - 0.01 * xdelta)
         #y = c(topsi, botsi, 0.5 * (topsi + botsi))
         #polygon(x = x, y = y, col = "#A2ACC9", border = "#A2ACC9")
         widths = (exproi$END - exproi$START) / xdelta
         labels = as.character(exproi$SCORE)
         index = strwidth(labels, units = "figure") > widths
         labels[index] = ""
         # leave only those for which score type is the same as for the first one
         index = exproi$scoreType != exproi$scoreType[[1]]
         labels[index] = ""
         # sort out duplicate records to output only 1 label - first occurance
         for (j in 1:(nexproi - 1)) {
            geneID = exproi$SEQ_ID[[j]]
            if (labels[[j]] != "")
               for (k in (j + 1):nexproi)
                  if (exproi$SEQ_ID[[k]] == geneID)
                     labels[[k]] = ""
         }
         text(x = 0.5 * (exproi$START + exproi$END), y = topsi + h$ticks, labels = labels, col = "black", adj = c(0.5, 0))
         axislabel = paste("score:", exproi$scoreType[[1]])
         text(x = xrange$min - 0.2 * xdelta, y = 0.5 * (topsi + botsi) + h$ticks, labels = axislabel, col = "black", adj = c(0, 0))
         text(x = xrange$min - 0.2 * xdelta, y = 0.5 * (topsi + botsi) - h$ticks, labels = names(expro)[[i]], col = "black", adj = c(0, 1), font = 2)
      }
   }
}

if(!isGeneric("plot.gff"))
   setGeneric("plot.gff", function(x, ...) standardGeneric("plot.gff"))

setMethod("plot.gff", signature(x = "list"),
   function(x, xrange, coco.tresholds, expro.tresholds, par.devstage, par.anatomy, h, nlevels = 8, coco.ramp = TRUE) {
      # coco.tresholds, expro.tresholds - named vectors of tresholds, names() - experiments
      # h - heights (see below)
      # 1. setting attributes to address fields by names
      if (missing(xrange)) {
         xrange <- c(10000, 100000)
      } else
         xrange = xrange[1:2]
      names(xrange) <- c("min", "max")
      xrange = as.list(xrange)
      xdelta = xrange$max - xrange$min
      if (missing(h))
         h = def.h()
      h = h[1:10]
      names(h) <- c("gap", "text", "coco", "cocoSep", "curve", "gene", "expro", "exproSep", "axis", "ticks")
      h = as.list(h)
      # 2. trim data to the range
      data = gffGetSubset(x, xrange)
      # 3. a la variable substitutions, sorting cocos, expros to match par.*
      if (!is.null(data$genomic_DNA)) {
         index = match(names(data$genomic_DNA), names(coco.tresholds))
         index = index[!is.na(index)]
         coco = data$genomic_DNA[index]
         ncoco = length(coco)
      }
      else
        ncoco = 0
      if (!is.null(data$gene))
         gene = data$gene
      if (!is.null(data$exon))
         exon = data$exon
      if (!is.null(data$expro)) {
         index = match(names(data$expro), names(expro.tresholds))
         index = index[!is.na(index)]
         expro = data$expro[index]
         nexpro = length(expro)
      }
      else
        nexpro = 0
      # 4. calculate plot height and element tops (p): see 'plot-mag.svg' for layout
      # nlevels - how many cocolevels one beneath the other
      ydelta = nlevels * ncoco * (h$coco + h$cocoSep) - h$cocoSep +            # cocos
               2 * (nexpro * (h$expro + h$text + h$exproSep) - h$exproSep) + # expro
               2 * (h$gene + 2 * h$text) +                                     # genes
               h$curve + h$xis                                                 # curve & axis
               8 * h$gap
      p = list()
      p$coco = ncoco * (nlevels * (h$coco + h$cocoSep) + h$gap) - h$cocoSep - h$gap
      p$up = numeric(3)
      yrange = numeric(2)
      p$up[[1]]   = p$coco + h$gap + h$text + h$gene
      p$up[[2]]   = p$up[[1]] + h$text + h$gap + h$curve
      p$up[[3]]   = p$up[[2]] + 2 * h$gap + nexpro * (h$expro + h$text + h$exproSep) - h$exproSep
      yrange[[2]] = p$up[[3]] + h$gap
      p$down = numeric(3)
      p$down[[1]] = h$gap + h$text
      p$down[[2]] = p$down[[1]] + h$gene + h$text + h$gap
      p$down[[3]] = p$down[[2]] + h$axis + h$gap
      p$down = -1 * p$down
      yrange[[1]] = p$down[[3]] - (nexpro * (h$expro + h$text + h$exproSep) - h$exproSep) - h$gap
      names(yrange) <- c("min", "max")
      yrange = as.list(yrange)
      # 5. create plot
      plot.new()
      par(mai = c(0, 0, 0, 0), cex = 0.6)
      xlim = as.numeric(xrange)
      xlim[[1]] = xlim[[1]] - 0.15 * xdelta
      xlim[[2]] = xlim[[2]] + 0.1 * xdelta
      plot.window(xlim = xlim, ylim = as.numeric(yrange))
      # 6. plot axis
      gffplot.axis(p$down[[2]], xrange, yrange, h)
      # 7. plot coco and coco expression curve
      if (exists("coco")) {
         gffplot.coco(coco, xrange, coco.tresholds, p$coco, h, nlevels, do.ramp = coco.ramp)
         gffplot.cocoCurve(coco[[1]], xrange, coco.tresholds, p$up[[2]], h)
      }
      # 8. plot genes and exons
      if (exists("gene")) {
         genesub = gene[gene$STRAND == "+", ]
         if (exists("exon"))
            exonsub = exon[exon$STRAND == "+", ]
         else
            exonsub = NULL
         if (exists("insitu"))
            insitusub = insitu[insitu$STRAND == "+", ]
         else
            insitusub = NULL
         gffplot.genesAndExons(genesub, exonsub, insitusub, "+", xrange, par.devstage, par.anatomy, p$up[[1]], h)
         genesub = gene[gene$STRAND == "-", ]
         if (exists("exon"))
            exonsub = exon[exon$STRAND == "-", ]
         else
            exonsub = NULL
         if (exists("insitu"))
            insitusub = insitu[insitu$STRAND == "-", ]
         else
            insitusub = NULL
         gffplot.genesAndExons(genesub, exonsub, insitusub, "-", xrange, par.devstage, par.anatomy, p$down[[1]], h)
      }
      # 9. plot expro
      if (exists("expro")) {
         gffplot.expro(expro, "+", xrange, expro.tresholds, p$up[[3]], h)
         gffplot.expro(expro, "-", xrange, expro.tresholds, p$down[[3]], h)
      }
      invisible(NULL)
   }
)

plot.gff.toFile <- function(x, file, type = "jpg", width = 1024, height = 768, use.GDD = FALSE, ...) {
   openWithGDD = FALSE
   if (use.GDD)
      if (require(GDD)) {
         GDD(file = file, type = type, width, height)
         openWithGDD = TRUE
      }
   if (!openWithGDD) {
      if (type == "jpg" || type == "jpeg")
         jpeg(file, width, height)
      else
         png(file, width, height)
   }
   plot.gff(x, ...)
   dev.off()
   invisible(NULL)
}
