##############################################################
## gff3Plotter - GFF3 file loader and preformatter          ##
## Parts of code based on <tilingArray> package - see marks ##
## Copyright: Oleg Sklyar 2005                              ##
## European Bioinformatics Institute, Cambs, UK             ##
## osklyar@ebi.ac.uk                                        ##
##############################################################

# for command line execution, please use
# R --vanilla --slave --args example.txt < batch.R > output

gffTypes <- c("gene", "exon", "insitu", "regulatory_region", "genomic_DNA", "expro")

##############################################################
## Returns a vector of <value>s for attribute <field> in    ##
## the character vector stored as <field>=<value> and by    ##
## default separated with ; or ;<space>                     ##
##############################################################
## Adapted from tilingArray::getAttributeField: regular     ##
## expressions are used insted of fixed=TRUE                ##
##############################################################
getAttr <- function (x, field, splitter = ";[[:space:]]?") {
   s = strsplit(x, split = splitter, extended = TRUE)
   sapply(s,
      function(atts) {
         a = strsplit(atts, split = "=", fixed = TRUE)
         m = match(field, sapply(a, "[", 1))
         if (!is.na(m))
            rv = a[[m]][2]
      else
            rv = as.character("")
         return(rv)
    }
   )
}

##############################################################
## Loads and Preformats GFF3 data                           ##
##############################################################
read.gff <- function(fileName, keepTypes = gffTypes) {
   cat("gffLoad# LOADING <", fileName, ">\n")
   data = read.table(fileName, sep= "\t", header= TRUE, comment.char= "#", na.strings= ".", as.is= TRUE, quote="")
  if(is.null(data)) return(NA)
   cat("gffLoad# Removing Extra Data\n")
   data = data[data$TYPE %in% keepTypes, ]
   cat("gffLoad# Splitting Types\n")
   data = split(data, data$TYPE)
   cat("done\n")
   if(!is.null(data$gene)) {
      cat("gffLoad# Splitting <gene> Attributes:")
      attr = data$gene$ATTRIBUTES
      data$gene$Name = getAttr(attr, "Name")
      cat(" Name-done")
      data$gene$ID = getAttr(attr, "ID")
      cat("; ID-done\n")
      cat("gffLoad# Removing ATTRIBUTES Field From <gene>: ")
      data$gene = data$gene[ , -9]
      cat("done\n")
   }
   # remove all data but tiling part if no genes (they do not have sense then)
   else
      data = data[names(data)=="genomic_DNA"]
   if(!is.null(data$genomic_DNA)) {
      cat("gffLoad# Splitting <genomic_DNA> Attributes:")
      attr = data$genomic_DNA$ATTRIBUTES
      data$genomic_DNA$expName = getAttr(attr, "expName")
      cat(" expNames-done")
      data$genomic_DNA$scoreType = getAttr(attr, "scoreType")
      cat("; scoreType-done")
      data$genomic_DNA$coeff = as.numeric(getAttr(attr, "coeff"))
      cat("; coeff-done")
      data$genomic_DNA$coeffType = getAttr(attr, "coeffType")
      cat("; coeffType-done")
      data$genomic_DNA$sticky = as.logical(getAttr(attr, "sticky"))
      cat("; sticky-done\n")
      cat("gffLoad# Removing ATTRIBUTES Field From <genomic_DNA>: ")
      data$genomic_DNA = data$genomic_DNA[ , -9]
      cat("done\n")
   }
   if(!is.null(data$expro)) {
      cat("gffLoad# Splitting <expro> Attributes:")
      attr = data$expro$ATTRIBUTES
      data$expro$expName = getAttr(attr, "expName")
      cat(" expNames-done")
      data$expro$scoreType = getAttr(attr, "scoreType")
      cat("; scoreType-done")
      data$expro$coeff = as.numeric(getAttr(attr, "coeff"))
      cat("; coeff-done")
      data$expro$coeffType = getAttr(attr, "coeffType")
      cat("; coeffType-done\n")
      cat("gffLoad# Removing ATTRIBUTES Field From <expro>: ")
      data$expro = data$expro[ , -9]
      cat("done\n")
   }
   if(!is.null(data$insitu)) {
      cat("gffLoad# Splitting <insitu> Attributes:")
      attr = data$insitu$ATTRIBUTES
      data$insitu$anatomy = getAttr(attr, "anatomy")
      cat(" anatomy-done")
      data$insitu$devStage = getAttr(attr, "devStage")
      cat("; devStage-done\n")
      cat("gffLoad# Removing ATTRIBUTES Field From <insitu>: ")
      data$insitu = data$insitu[ , -9]
      cat("done\n")
   }
   cat("gffLoad# Removing ATTRIBUTES Field From Other Types:")
   if(!is.null(data$exon)) {
      data$exon = data$exon[ , -9]
      cat(" exon-done;")
   }
   if(!is.null(data$regulatory_region)) {
      data$regulatory_region = data$regulatory_region[ , -9]
      cat(" regulatory_region-done")
   }
   cat(" complete\n")
   cat("gffLoad# Sorting START of genomic_DNA: ")
   if(!is.null(data$genomic_DNA))
      if(length(data$genomic_DNA$START) > 0)
         data$genomic_DNA = data$genomic_DNA[order(data$genomic_DNA$START), ]
   cat(" complete\n")
   if(!is.null(data$gene) && !is.null(data$expro)) {
      cat("gffLoad# Removing <expro> Records matching no Genes")
      data$expro = data$expro[which(data$expro$SEQ_ID %in% data$gene$ID), ]
      cat(" - done\n")
      cat("gffLoad# Updating <expro> START,END,STRAND Fields from <gene>")
      index = match(data$expro$SEQ_ID, data$gene$ID)
      data$expro[1:length(data$expro$SEQ_ID), 4:5] <- data$gene[index, 4:5]
      data$expro[1:length(data$expro$SEQ_ID), 7] <- data$gene[index, 7]
      cat(" - done\n")
   }
   if(!is.null(data$expro))
      if(!is.null(data$expro$expName)) {
         cat("gffLoad# Splitting <expro> by <expName>")
         #cat(data$expro$SCORE)
         data$expro = split(data$expro, data$expro$expName)
         #cat(data$expro[[1]]$SCORE)
         cat(" - done\n")
      }
   if(!is.null(data$gene) && !is.null(data$insitu)) {
      cat("gffLoad# Removing <insitu> Records matching no Genes")
      data$insitu = data$insitu[which(data$insitu$SEQ_ID %in% data$gene$ID), ]
      cat(" - done\n")
      cat("gffLoad# Updating <insitu> START,END,STRAND Fields from <gene>")
      index = match(data$insitu$SEQ_ID, data$gene$ID)
      data$insitu[1:length(data$insitu$SEQ_ID), 4:7] <- data$gene[index, 4:7]
      cat(" - done\n")
   }
   if(!is.null(data$genomic_DNA))
      if(!is.null(data$genomic_DNA$expName)) {
         cat("gffLoad# Splitting <genomic_DNA> by <expName>")
         data$genomic_DNA = split(data$genomic_DNA, data$genomic_DNA$expName)
         cat(" - done\n")
      }
  cat("gffLoad# LOAD COMPLETE\n")
  #cat(data$expro[[1]]$SCORE)
  invisible(data)
}

##############################################################
## Returns a subset of structured GFF3 data in              ##
## min=range[[1]] and max=range[[2]] ##                     ##
##############################################################
gffGetSubset <- function(data, range) {
  newData = vector(mode= "list", length= length(data))
  names(newData) <- names(data)
  if(!is.null(data$gene))
    newData$gene = data$gene[data$gene$START <= range[[2]] & data$gene$END >= range[[1]], ]
  if(!is.null(data$exon))
    newData$exon = data$exon[data$exon$START <= range[[2]] & data$exon$END >= range[[1]], ]
  if(!is.null(data$regulatory_region))
    newData$regulatory_region = data$regulatory_region[data$regulatory_region$START <= range[[2]] & data$regulatory_region$END >= range[[1]], ]
  if(!is.null(data$insitu))
    newData$insitu = data$insitu[data$insitu$START <= range[[2]] & data$insitu$END >= range[[1]], ]
  if(!is.null(data$genomic_DNA)) {
   newData$genomic_DNA = vector(mode = "list", length = length(data$genomic_DNA))
   names(newData$genomic_DNA) = names(data$genomic_DNA)
   for(i in 1:length(data$genomic_DNA))
      newData$genomic_DNA[[i]] = data$genomic_DNA[[i]][data$genomic_DNA[[i]]$START <= range[[2]] & data$genomic_DNA[[i]]$END >= range[[1]], ]
   }
  if(!is.null(data$expro)) {
   newData$expro = vector(mode = "list", length = length(data$expro))
   names(newData$expro) = names(data$expro)
   for(i in 1:length(data$expro))
      newData$expro[[i]] = data$expro[[i]][data$expro[[i]]$START <= range[[2]] & data$expro[[i]]$END >= range[[1]], ]
   }
  return(newData)
}

##############################################################
## Extracts a subset of a data frame, in which attributes   ##
## a field 'attribute' occurs with value 'value'            ##
##############################################################
## TODO: remove - PROBABLY UNUSED
gffGetRecordsWith <- function(dataSubset, attribute, value) {
   doGet <- function(x) return(x[attribute] == value)
   invisible(dataSubset[unlist(lapply(dataSubset$ATTRIBUTES, doGet)), ])
}
